import json
import base64
import time
import os
import uuid
import hashlib
import random
import string
from urllib.parse import urlparse, parse_qs, quote, urlencode
from urllib.request import urlopen, Request, build_opener, HTTPCookieProcessor
from urllib.error import HTTPError
from http.cookiejar import CookieJar
import datetime
import re

class TokenNotFound(Exception):
    pass

class TokenRenewFailed(Exception):
    pass

class RequestFailed(Exception):
    pass



class APIParser():
    def strptime(str, format):
        try:
            return datetime.datetime.strptime(str, format)
        except TypeError:
            return datetime.datetime(*(time.strptime(str, format)[0:6]))

    class Season():
        def __init__(self, id, title, numberOfEpisodes, seasonType, year=None, month=None, no=None):
            self.id = id
            self.title = title
            self.numberOfEpisodes = numberOfEpisodes
            self.seasonType = seasonType
            self.year = year
            self.month = month
            self.no = no
        def __str__(self):
            return f"ID: {self.id}, Title: {self.title}, numberOfEpisodes {self.numberOfEpisodes}"

    class RecommendationOverview():
        def __init__(self, id, title):
            self.id = id
            self.title = title

        def __str__(self):
            return f"ID: {self.id}, Title: {self.title}"

    class Event():
        def __init__(self, id, title, tier, description, startTimestamp, landscapeImg, portraitImage):
            self.id = id
            self.title = title
            self.tier = tier
            self.startTimestamp = startTimestamp
            self.description = description
            self.landscapeImg = landscapeImg
            self.portraitImage = portraitImage
        def __str__(self):
            return f"ID: {self.id}, Title: {self.title}, Description: {self.description}, Start Timestamp: {self.startTimestamp}, Tier {self.tier}, Landscape Image: {self.landscapeImg}, Portrait Image: {self.portraitImage}"

    class FormatOverview():
        def __init__(self, id, title, tier, formatType, landscapeImg, portraitImage, description=None):
            self.id = id
            self.title = title
            self.tier = tier
            self.formatType = formatType
            self.description = description
            self.landscapeImg = landscapeImg
            self.portraitImage = portraitImage
        def __str__(self):
            return f"ID: {self.id}, Title: {self.title}, Description: {self.description}, Tier {self.tier}, Landscape Image: {self.landscapeImg}, Portrait Image: {self.portraitImage}"

    class FormatDescription():
        def __init__(self, id, title, tier, description, productionYear, productionCountries, numberOfSeasons, numberOfEpisodes, genres, seasons, episodeNo, season, duration, ageRating, landscapeImg, portraitImage, airDate=None):
            self.id = id
            self.title = title
            self.tier = tier
            self.description =description
            self.productionYear = productionYear
            self.productionCountries = productionCountries
            self.numberOfSeasons = numberOfSeasons
            self.numberOfEpisodes = numberOfEpisodes
            self.genres = genres
            self.seasons = seasons
            self.episodeNo = episodeNo
            self.season = season
            self.duration = duration
            self.ageRating = ageRating
            self.landscapeImg = landscapeImg
            self.portraitImage = portraitImage
            self.airDate = airDate
        def __str__(self):
            return f"ID: {self.id}, Title: {self.title}, Tier {self.tier}, Description: {self.description}, Production Year: {self.productionYear}, " \
                   f"Production Countries: {self.productionCountries}, Number of Episodes: {self.numberOfEpisodes}, Number of Seasons: {self.numberOfSeasons}, " \
                   f"Genres: {self.genres}, Seasons: {self.seasons}, Landscape Image: {self.landscapeImg}, Portrait Image: {self.portraitImage}"

    @staticmethod
    def getElement(json, path : list, allowNone : bool = False ):
        while json is not None and len(path) > 0:
            json = json.get(path[0])
            del path[0]
        if not allowNone and json is None:
            raise Exception("Element not found")
        return json

    def parseFormatDetailed(json) -> FormatDescription:
        id = APIParser.getElement(json, ["id"])
        title = APIParser.getElement(json, ["title"])
        tier = APIParser.getElement(json, ["tier"], True)
        description = APIParser.getElement(json, ["emptyFormatText"], True)
        if description is None:
            description = APIParser.getElement(json, ["descriptionV2"], True)
        productionYear = APIParser.getElement(json, ["productionYear"], True)
        productionCountries = APIParser.getElement(json, ["productionCountries"], True)
        numberOfSeasons = APIParser.getElement(json, ["numberOfSeasons"], True)
        numberOfEpisodes = APIParser.getElement(json, ["numberOfEpisodes"], True)
        episodeNo = APIParser.getElement(json, ["number"], True)
        duration = APIParser.getElement(json, ["durationInSecondsV2"], True)
        ageRating = APIParser.getElement(json, ["ageRating"], True)
        genres = APIParser.getElement(json, ["genres"], True)
        landscapeImg = APIParser.getElement(json, ["watchImages", "artworkLandscape", "absoluteUri"], True)
        portraitImage = APIParser.getElement(json, ["watchImages", "plainLandscape", "absoluteUri"], True)
        if portraitImage is None:
            portraitImage = APIParser.getElement(json, ["watchImages", "artworkPortrait", "absoluteUri"], True)
        if landscapeImg is None:
            landscapeImg = APIParser.getElement(json, ["watchImages", "default", "absoluteUri"], True)
        seasons = APIParser.parseSeasons(APIParser.getElement(json, ["seasons"], True))
        season = APIParser.getElement(json, ["episodeSeason", "season", "ordinal"], True)
        return APIParser.FormatDescription(id, title, tier, description, productionYear, productionCountries, numberOfSeasons, numberOfEpisodes, genres, seasons, episodeNo, season, duration, ageRating, landscapeImg, portraitImage)

    def parseFormatOverview(json) -> FormatOverview:
        id = APIParser.getElement(json, ["id"])
        title = APIParser.getElement(json, ["title"])
        tier = APIParser.getElement(json, ["tier"], True)
        formatType = APIParser.getElement(json, ["__typename"])
        landscapeImg = APIParser.getElement(json, ["watchImages", "artworkLandscape", "absoluteUri"], True)
        if landscapeImg is None:
            landscapeImg = APIParser.getElement(json, ["images", "artworkLandscape", "url"], True)
        portraitImage = APIParser.getElement(json, ["watchImages", "artworkPortrait", "absoluteUri"], True)
        if portraitImage is None:
            portraitImage = APIParser.getElement(json, ["images", "artworkPortrait", "url"], True)
        description = APIParser.getElement(json, ["urlData", "seo", "description"], True)
        return APIParser.FormatOverview(id, title, tier, formatType, landscapeImg, portraitImage, description)

    def parseEvent(json) -> Event:
        id = APIParser.getElement(json, ["id"])
        title = APIParser.getElement(json, ["title"])
        tier = APIParser.getElement(json, ["tier"], True)
        description = APIParser.getElement(json, ["urlData", "seo", "description"], True)
        streamStart = APIParser.getElement(json, ["streamStart"], True)
        startTimestamp = 0
        if streamStart is not None:
            cleanTimeStr = str(streamStart).split('.')[0]
            display = '%Y-%m-%dT%H:%M:%S'
            dateTimeObj = APIParser.strptime(cleanTimeStr, display)
            startTimestamp = time.mktime(dateTimeObj.timetuple())
        landscapeImg = APIParser.getElement(json, ["watchImages", "artworkLandscape", "absoluteUri"], True)
        if landscapeImg is None:
            landscapeImg = APIParser.getElement(json, ["images", "artworkLandscape", "url"], True)
        portraitImage = APIParser.getElement(json, ["watchImages", "artworkPortrait", "absoluteUri"], True)
        if portraitImage is None:
            portraitImage = APIParser.getElement(json, ["images", "artworkPortrait", "url"], True)

        return APIParser.Event(id, title, tier, description, startTimestamp, landscapeImg, portraitImage)

    def parseEpisodes(json) -> list:
        elements = APIParser.getElement(json, ["data", "season", "episodes"])
        episodes = []
        for el in elements:
            episodes.append(APIParser.parseFormatDetailed(el))
        return episodes

    def parseSeasons(json) -> list:
        seasons = []
        if json is None:
            return []
        for season in json:
            id = APIParser.getElement(season, ["id"])
            numberOfEpisodes = APIParser.getElement(season, ["numberOfEpisodes"], True)
            seasonType = APIParser.getElement(season, ["seasonType"], True)
            month = APIParser.getElement(season, ["month"], True)
            year = APIParser.getElement(season, ["year"], True)
            ordinal = APIParser.getElement(season, ["ordinal"], True)

            if seasonType == "ANNUAL":
                title = f'{month:02d}.{year:04d}'
            else:
                title = f'Staffel {ordinal}'

            seasons.append(APIParser.Season(id, title, numberOfEpisodes, seasonType, year, month, ordinal))
        return seasons

    def parseFormats(json) -> list:
        formats = []
        for found in json:
            formats.append(APIParser.parseFormatOverview(found))
        return formats

    def buildLiveDescription(title, description, startDate=None, endDate=None):
        descriptionText = description if description is not None else ""
        titleText = title if title is not None else ""
        startTimestamp = None
        endTimestamp = None
        if startDate is not None:
            cleanTimeStr = str(startDate).split('.')[0]
            display = '%Y-%m-%dT%H:%M:%S'
            dateTimeObj = APIParser.strptime(cleanTimeStr, display).replace(tzinfo=datetime.timezone.utc).astimezone(tz=None)
            startTimestamp = dateTimeObj.strftime('%H:%M')
        if endDate is not None:
            cleanTimeStr = str(endDate).split('.')[0]
            display = '%Y-%m-%dT%H:%M:%S'
            dateTimeObj = APIParser.strptime(cleanTimeStr, display).replace(tzinfo=datetime.timezone.utc).astimezone(tz=None)
            endTimestamp = dateTimeObj.strftime('%H:%M')
        if startTimestamp is not None and endTimestamp is not None:
            return f'{startTimestamp}-{endTimestamp}: [B]{titleText}[/B][CR]{descriptionText}[CR]'
        elif endTimestamp is not None:
            return f'bis {endTimestamp}: [B]{titleText}[/B][CR]{descriptionText}[CR]'
        elif startTimestamp is not None:
            return f'ab {endTimestamp}: [B]{titleText}[/B][CR]{descriptionText}[CR]'
        return f'[B]{titleText}[/B][CR]{descriptionText}[CR]'

    def parseLiveChannel(json) -> list:
        id = APIParser.getElement(json, ["id"])
        epgElement = APIParser.getElement(json, ["epgsFromNow"])
        epgElementNow = None
        if epgElement is not None and len(epgElement) > 0:
            epgElementNow = epgElement[0]
        onAirTitle = APIParser.getElement(epgElementNow, ["title"], True)
        title = f'[B]{APIParser.getElement(json, ["name"])}[/B]'
        if onAirTitle is not None:
            title += f' - {onAirTitle}'
        description = APIParser.buildLiveDescription(onAirTitle,
                                                     APIParser.getElement(epgElementNow, ["description"], True),
                                                     endDate=APIParser.getElement(epgElementNow, ["endPublished"], True))
        if epgElement is not None and len(epgElement) > 1:
            epgElementNext = epgElement[1]
            description += APIParser.buildLiveDescription(APIParser.getElement(epgElementNext, ["title"], True),
                                                     APIParser.getElement(epgElementNext, ["description"], True),
                                                     APIParser.getElement(epgElementNext, ["startPublished"], True),
                                                     APIParser.getElement(epgElementNext, ["endPublished"], True))

        tier = "PREMIUM"

        return APIParser.FormatOverview(id, title, tier, "LIVE", None, APIParser.getElement(epgElementNow, ["imageUrl"], True), description)

    def parseLiveChannels(json) -> list:
        elements = APIParser.getElement(json, ["data", "liveTvStations"])
        formats = []
        for el in elements:
            formats.append(APIParser.parseLiveChannel(el))
        return formats

    def parseEvents(json) -> list:
        teaserRows = APIParser.getElement(json, ["data", "liveEventsOverview", "teaserRows"])
        events = []
        for row in teaserRows:
            for event in APIParser.getElement(row, ["events"]):
                events.append(APIParser.parseEvent(event))
        return events

class API(): # EDITED = 26.06.2024 @realvito
    USERAGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:122.0) Gecko/20100101 Firefox/122.0"
    ENDPOINTS = {
        'Format' : '3d9aefd9bbfb25a66cea7bb67ca67b191959d805bfd8203eaaaad53a0b6e4a36',
        'SeasonWithFormatAndEpisodes' : '11963760dd159c9ba6d0c1061833c01aed86271f0f9c36586e5501f5e57d4fcd',
        'MRE' : '508c672e558432532a9fe25710ee2d27efb9db0b4325c3adcef238f23d2974ec',
        'Recommendations' : '3ac5d68d7b3d4678d77e32212887dcabcc01fdbfabff7c6918fc542e7a633e1f',
        'OverviewPage' : '28aad4e992bb63330bfcd40a6906af3119d8a2612fa9fd28dae9c19127e247ca',
        'LiveTvStations' : '845cf56a2a78110a0f978c1a2af2bc7f9a1c937d0f324ffaf852a9a4414c8485',
        'LiveEventsOverviewPage' : '77a8f26d5de76daf801fcd7ae54bebd0ecabbeeb3ecb90889bc4a2e5590b7d20',
        'MovieDetail' : 'b1c360212cc518ddca2b8377813a54fa918ca424c08086204b7bf7d6ef626ac4',
        'WatchPlayerConfigV3' : 'fea0311fb572b6fded60c5a1a9d652f97f55d182bc4cedbdad676354a8d2797c',
        'Search' : '9b4d8cd668c4aad6e7dd3eb4b0ce6137159f76da1c68784936a7b72635025f0a',
        'ExploreWidgetWatch' : '6fa3a96460211475c47cffe605c204eb733d6605ff7db57c9eaea4617188f3bd',
        'addFavorite' : '22cb1ccd31a2cbe37ed6b3483bd3026edb5283d4a8d5fa4d3b7e8807d56af306',
        'deleteFavorite' : 'ee3b38bdbc9f4e723db702819cd955580cdb530d8574e80c3f75852e74ec7937',
        'GetFavoriteIds' : '77e8f3ee705a86812515ada975e1b6037f73e448185dd426777875bbc0511801',
        'GetWatchFavorites' : 'd3771a9f6c262ddfaf38ea86d174920687ca709caa5dd210d35667a95a1f9724',
        'EpisodeDetail' : '04e2f59b9c750df1137f9946258c17686cd4447511383ef05fcf699183f449b8'
    }
    APIBASEURL = base64.b64decode(b'aHR0cHM6Ly9jZG4uZ2F0ZXdheS5ub3ctcGx1cy1wcm9kLmF3cy1jYmMuY2xvdWQvZ3JhcGhxbA==').decode()
    AUTHBASEURL = base64.b64decode(b'aHR0cHM6Ly9hdXRoLnJ0bC5kZS9hdXRoL3JlYWxtcy9ydGxwbHVzL3Byb3RvY29sL29wZW5pZC1jb25uZWN0').decode()

    @staticmethod
    def quoteUrl(url : str):
        queryPos = url.find('?') + 1
        if queryPos > 0:
            parameters = url[queryPos:].split('&')
            baseURL = url[:queryPos]
            valuePairs = []
            for param in parameters:
                valuePos = param.find('=')
                valuePairs.append(param[:valuePos] + "=" + quote(param[valuePos+1:], safe=""))
            url = baseURL + "&".join(valuePairs)
        return url

    def postRequest(self, url : str, postData : dict=None, jsonData : json = None, token=None, client=False, opener=None, getUrl=False) -> str:
        headers = {
            'User-Agent': API.USERAGENT,
            'Referer' :  base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode(),
            'Origin' :  base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode(),
            'Content-Type' : "application/x-www-form-urlencoded"
        }
        if (postData is None and jsonData is None) or (postData is not None and jsonData is not None):
            return None
        if token is not None:
            headers["Authorization"] = f"Bearer {token}"
        if client:
            headers[base64.b64decode(b'cnRscGx1cy1jbGllbnQtSWQ=').decode()] = base64.b64decode(b'cmNpOnJ0bHBsdXM6d2Vi').decode()
            headers[base64.b64decode(b'cnRscGx1cy1jbGllbnQtVmVyc2lvbg==').decode()] = self.clientVersion
        if postData is not None:
            postData = urlencode(postData)
        if jsonData is not None:
            headers['Content-Type'] = "application/json; charset=utf-8"
            postData = json.dumps(jsonData)

        headers['Content-Length'] = len(postData)
        req = Request(self.quoteUrl(url), headers=headers)
        try:
            if opener is not None:
                query = opener.open(req, postData.encode('utf-8'))
            else:
                query = urlopen(req, postData.encode('utf-8'))

            if getUrl:
                res = query.geturl()
            else:
                res = query.read()
                print(res)
        except HTTPError as e:
        #if res.status_code != 200:
            print(e.code)
            print(e.read())
            raise RequestFailed("Invalid Response")
        return res

    def getRequest(self, url: str, token=None, client=True, opener=None) -> str:
        headers = {
            'User-Agent': API.USERAGENT,
            'Referer' :  base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode(),
            'Origin' :  base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode(),
        }
        if token is not None:
            headers["Authorization"] = f"Bearer {token}"
        if client:
            headers[base64.b64decode(b'cnRscGx1cy1jbGllbnQtSWQ=').decode()] = base64.b64decode(b'cmNpOnJ0bHBsdXM6d2Vi').decode()
            headers[base64.b64decode(b'cnRscGx1cy1jbGllbnQtVmVyc2lvbg==').decode()] = self.clientVersion

        req = Request(self.quoteUrl(url), headers=headers)
        try:
            if opener is not None:
                query = opener.open(req)
            else:
                query = urlopen(req)
            res = query.read()
            print(res)
        except HTTPError as e:
        #if res.status_code != 200:
            print(e.code)
            print(e.read())
            raise Exception("Invalid Response")
        return res

    @staticmethod
    def isTokenValid(token : str) -> bool:
        valid = True
        try:
            base64Parts = token.split(".")
            token = "%s==" % base64Parts[1]
            userData = json.loads(base64.b64decode(token))

            if "licenceEndDate" in userData:
                licenceEndDate = userData["licenceEndDate"].split("+")[4]
                try:
                    display = '%Y-%m-%dT%H:%M:%S'
                    licenceEndDateTS = APIParser.strptime(licenceEndDate, display)
                    licenceEndDateTS = time.mktime(licenceEndDateTS)
                    if licenceEndDateTS < (time.time() + 60):
                        valid = False
                except:
                    valid = False
            if (valid and "exp" in userData and
                userData["exp"] > (time.time() + 60)):
                valid = True
            else:
                valid = False
        except:
            valid = False
        return valid

    @staticmethod
    def parseTokenConfig(config):
        postData = {}
        for pair in config.split(","):
            splittedPair = pair.split(":")
            postData[splittedPair[0]] = splittedPair[1].replace('"','')
        return postData

    @staticmethod
    def getConfig():
        baseEndPoint = base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode()
        endPoint = baseEndPoint
        headers = {"User-Agent": API.USERAGENT}
        req = Request(API.quoteUrl(endPoint), headers=headers)
        query = urlopen(req)
        res = query.read().decode("UTF-8")
        try:
            jsName = re.findall(r'<script src="(main[A-z0-9\-\.]+\.js)"', res, re.S)[-1]
        except:
            return None
        endPoint = baseEndPoint + jsName
        req = Request(API.quoteUrl(endPoint), headers=headers)
        query = urlopen(req)
        res = query.read().decode("UTF-8")
        return res

    @staticmethod
    def getClientVersion():
        baseEndPoint = base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS9hc3NldHMvY29uZmlnL2NvbmZpZy5qc29u=').decode()
        endPoint = baseEndPoint
        headers = {"User-Agent": API.USERAGENT}
        req = Request(API.quoteUrl(endPoint), headers=headers)
        query = urlopen(req)
        jsonData = json.loads(query.read())
        try:
            return jsonData["version"]
        except:
            return None

    @staticmethod
    def getClientID() -> str:
        text = API.getConfig()
        m = re.search(r'clientId:"([^"]+)"', text)
        if m:
            return m.group(1)
        return None

    def receiveToken(self) -> str:
        text = API.getConfig()
        m = re.search(r'anonymousCredentials:{([^}]+)}', text)
        if m:
            authURL = API.AUTHBASEURL + "/token"
            postData = API.parseTokenConfig(m.group(1))
            try:
                res = self.postRequest(authURL, postData=postData)
                jsonData = json.loads(res)
                token = jsonData.get("access_token")
            except RequestFailed:
                return None
            return token
        return None

    def renewToken(self):
        authURL = API.AUTHBASEURL + "/token?ngsw-bypass="
        postData = {
                "grant_type": "refresh_token",
                "refresh_token": self.refreshToken,
                "client_id": self.clientID
        }
        try:
            res = self.postRequest(authURL, postData=postData)
            jsonData = json.loads(res)
            token = jsonData.get("access_token")
            refresh_token = jsonData.get("refresh_token")
            return token, refresh_token
        except RequestFailed:
            pass
        return None, None

    def getToken(self):
        return self.token

    def getRefreshToken(self):
        return self.refreshToken

    def __init__(self, token=None, refresh_token=None, clientID=None, clientVersion=None, fetchPremiere=False):
        self.refreshToken = refresh_token
        self.token = None
        self.fetchPremiere = fetchPremiere
        if clientID is None:
            self.clientID = API.getClientID()
        else:
            self.clientID = clientID
        if clientVersion is None:
            self.clientVersion = API.getClientVersion()
        else:
            self.clientVersion = clientVersion
        if token is None or token == "":
            token = self.receiveToken()
            if self.isTokenValid(token):
                self.token = token
        else:
            if self.isTokenValid(token):
                self.token = token
            elif self.refreshToken is not None and self.refreshToken != "":
                if self.isTokenValid(self.refreshToken):
                    self.token, self.refreshToken = self.renewToken()
                    if self.token is None or self.refreshToken is None:
                        raise TokenRenewFailed()
                else:
                    raise TokenRenewFailed()
            else:
                self.token = self.receiveToken()
        if self.token is None:
            raise TokenNotFound()

    class Login():
        def __init__(self, api):
            self.api = api
            code_verifier = ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(96))
            hashed = hashlib.sha256(code_verifier.encode('ascii')).digest()
            encoded = base64.urlsafe_b64encode(hashed)
            code_challenge = encoded.decode('ascii')[:-1]
            state = uuid.uuid4()
            nonce = uuid.uuid4()
            reqURL = f'{API.AUTHBASEURL}/auth?client_id={self.api.clientID}&redirect_uri=https://{base64.b64decode(b"cGx1cy5ydGwuZGU==").decode()}/&state={state}&response_mode=query&response_type=code&scope=openid email&nonce={nonce}&prompt=login&code_challenge={code_challenge}&code_challenge_method=S256'
            cj = CookieJar()
            self.opener = build_opener(HTTPCookieProcessor(cj))
            res = self.api.getRequest(reqURL, opener=self.opener)
            pattern = '<form id="rtlplus-form-login" action="([^"]*)"'
            reqURL = re.findall(pattern, res.decode('UTF-8'))[0]
            import html
            self.loginURL = html.unescape(reqURL)
            self.code_verifier = code_verifier
            self.access_token = None
            self.refresh_token = None


        def sendLogin(self, username, password):
            if self.loginURL is None or self.api.clientID is None or self.code_verifier is None:
                return False
            data ={
                "credentialId": "",
                "rememberMe": "on",
                "username": username,
                "password": password
            }
            try:
                url = self.api.postRequest(self.loginURL, postData=data, opener=self.opener, getUrl=True)
            except:
                return False
            code = parse_qs(url)['code'][0]
            data = {
                "code": code,
                "grant_type": "authorization_code",
                "client_id": self.api.clientID,
                "redirect_uri": base64.b64decode(b'aHR0cHM6Ly9wbHVzLnJ0bC5kZS8=').decode(),
                "code_verifier": self.code_verifier
            }

            authURL = f"{API.AUTHBASEURL}/token?ngsw-bypass="
            try:
                res = self.api.postRequest(authURL, postData=data, opener=self.opener)
            except:
                return False

            try:
                data = json.loads(res)
                self.access_token = data["access_token"]
                self.refresh_token = data["refresh_token"]
                self.id_token = data["id_token"]
                self.api.token = self.access_token
            except:
                print(res.text)
                return False
            return True

        def getAccessToken(self):
            return self.access_token

        def getRefreshToken(self):
            return self.refresh_token

    def getAirDates(self, episodes):
        def fetchAirDate(self, episodeTpl):
            episode = episodeTpl[1]
            idx = episodeTpl[0]
            epId = episode.id
            requestURL = f'{API.APIBASEURL}?operationName=EpisodeDetail&variables={{"episodeId": "{epId}"}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["EpisodeDetail"]}"}}}}'
            res = self.getRequest(requestURL, self.token)
            jsonData = json.loads(res)
            airDate = APIParser.getElement(jsonData, ["data", "episode", "firstAiredDate"], True)
            if airDate is not None:
                display = '%Y-%m-%dT%H:%M:%S'
                cleanTimeStr = str(airDate).split('.')[0]
                dateTimeObj = APIParser.strptime(cleanTimeStr, display)
                airDate = time.mktime(dateTimeObj.timetuple())
            return idx, airDate
        import concurrent.futures
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            fetchers = {executor.submit(fetchAirDate, self, episode): episode for episode in enumerate(episodes)}
            for future in concurrent.futures.as_completed(fetchers):
                try:
                    data = future.result()
                except Exception as exc:
                    print('%r generated an exception: %s' % (fetchers[future], exc))
                else:
                    episodes[data[0]].airDate = data[1]
        return episodes

    def getFormat(self, id):
        requestURL = f'{API.APIBASEURL}?operationName=Format&variables={{"id": "{id}" }}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["Format"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        format = APIParser.parseFormatDetailed(APIParser.getElement(jsonData, ["data", "format"]))
        return format

    def getSeason(self, seasonId):
        offset = 0
        limit = 5
        episodes = []
        while True:
            requestURL = f'{API.APIBASEURL}?operationName=SeasonWithFormatAndEpisodes&variables={{"seasonId": "{seasonId}","offset":{offset},"limit":{limit} }}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["SeasonWithFormatAndEpisodes"]}"}}}}'
            res = self.getRequest(requestURL, self.token)
            jsonData = json.loads(res)
            episodes += APIParser.parseEpisodes(jsonData)
            totalItems = APIParser.getElement(jsonData, ["data", "season", "numberOfEpisodes"])
            offset += limit
            if offset >= totalItems:
                break
        if self.fetchPremiere:
            episodes = self.getAirDates(episodes)

        return episodes

    def getMovie(self, id):
        requestURL = f'{API.APIBASEURL}?operationName=MovieDetail&variables={{"id": "{id}" }}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["MovieDetail"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        format = APIParser.parseFormatDetailed(APIParser.getElement(jsonData, ["data", "movie"]))
        return format

    def _parseManifest(self, jsonData):
        sources = APIParser.getElement(jsonData, ["sources"])
        playbackURL = None
        if sources is not None:
            for source in sources:
                playbackURL = APIParser.getElement(source, ["url"])
                isYospace = APIParser.getElement(source, ["isYospace"])
                if isYospace is None or isYospace is False:
                    break
        if playbackURL is None:
            playbackURL = APIParser.getElement(jsonData, ["videoUrl"])
        licenseURL = None
        for license in APIParser.getElement(jsonData, ["licenses"]):
            licenseType = APIParser.getElement(license, ["type"])
            if licenseType == "WIDEVINE":
                licenseURL = APIParser.getElement(license, ["licenseUrl"])
        return playbackURL, licenseURL

    def getPlaybackURL(self, id, hdEnabled=True):
        requestURL = f'{API.APIBASEURL}?operationName=WatchPlayerConfigV3&variables={{"platform":"WEB","id": "{id}" }}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["WatchPlayerConfigV3"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        playoutVariants = APIParser.getElement(jsonData, ["data", "watchPlayerConfigV3", "playoutVariants"])
        playbackURL = None
        licenseURL = None
        for variant in playoutVariants:
            manifestType = APIParser.getElement(variant, ["type"])
            if (hdEnabled or playbackURL is None) and manifestType == "dashhd":
                playbackURL, licenseURL = self._parseManifest(variant)
                if hdEnabled:
                    break
            if manifestType == "dashsd":
                playbackURL, licenseURL = self._parseManifest(variant)
                if not hdEnabled:
                    break
        return playbackURL, licenseURL

    def getOverview(self, elementType):
        limit = 48
        offset = 0
        apiParams = {
            'APIBASEURL' : API.APIBASEURL,
            'offset' : offset,
            'limit' : limit,
            'elementType' : elementType,
            'endpoint' : API.ENDPOINTS["OverviewPage"]
        }
        requestURL = '{APIBASEURL}?operationName=OverviewPage&variables={{"pagination":{{"offset":{offset}, "limit":{limit}}},"filter":{{"elementType":"{elementType}"}}}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{endpoint}"}}}}'
        res = self.getRequest(requestURL.format(**apiParams), self.token)
        jsonData = json.loads(res)
        formats = APIParser.parseFormats(APIParser.getElement(jsonData, ["data", "watchOverviewPage", "elements"]))
        totalItems = APIParser.getElement(jsonData, ["data", "watchOverviewPage", "pageInfo", "totalCount"])
        apiParams['offset'] += limit
        apiParams['limit'] = totalItems - limit
        res = self.getRequest(requestURL.format(**apiParams), self.token)
        jsonData = json.loads(res)
        formats += APIParser.parseFormats(APIParser.getElement(jsonData, ["data", "watchOverviewPage", "elements"]))
        return formats

    def getLive(self):
        requestURL = f'{API.APIBASEURL}?operationName=LiveTvStations&variables={{"epgCount":4,"filter":{{"channelTypes":["BROADCAST"]}}}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["LiveTvStations"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        format = APIParser.parseLiveChannels(jsonData)
        return format

    def getEvents(self):
        requestURL = f'{API.APIBASEURL}?operationName=LiveEventsOverviewPage&variables={{}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["LiveEventsOverviewPage"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        format = APIParser.parseEvents(jsonData)
        return format

    def search(self, query):
        requestURL = f'{API.APIBASEURL}?operationName=Search&variables={{"input":{{"scopes":["WATCH","MUSIC","PODCAST","AUDIOBOOKS","MAGAZINE"],"term":"{query}"}}}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["Search"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)

        for result in APIParser.getElement(jsonData, ["data", "search", "result"]):
            key = APIParser.getElement(result, ["titleKey"])
            if key == "WATCH":
                return APIParser.parseFormats(APIParser.getElement(result, ["items"]))

    def modiyWishlist(self, itemID, action):
        data = {
            "operationName": action,
            "variables": {
                "id": itemID
            },
            "extensions": {
                "persistedQuery": {
                    "version":1,
                    "sha256Hash": API.ENDPOINTS[action]
                    }
                }
            }
        try:
            self.postRequest(API.APIBASEURL, jsonData=data, token=self.token, client=True)
            return True
        except:
            return False

    def getRecommendationOverview(self, offset=0, take=15, cacheFile=None):
        requestURL = f'{API.APIBASEURL}?operationName=ExploreWidgetWatch&variables={{"area":"home","offset":{offset},"take":{take}}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["ExploreWidgetWatch"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        recommendations = []
        try:
            jsonData = json.loads(res)
            for result in APIParser.getElement(jsonData, ["data", "editorialView", "widgets", "items"]):
                widgetName = APIParser.getElement(result, ["__typename"])
                if widgetName in ["TeaserWidget", "SavedFavoriteWidget", "WidgetContinue"]:
                    elements = APIParser.getElement(result, ["elements"])
                    if len(elements) > 0:
                        id = APIParser.getElement(result, ["id"])
                        title = APIParser.getElement(result, ["title"])
                        recommendations.append(APIParser.RecommendationOverview(id, title))
            if cacheFile is not None:
                try:
                    os.remove(cacheFile)
                except OSError:
                    pass
                with open(cacheFile, "w", encoding='utf-8') as f:
                    f.write(res.decode())
        except:
            pass
        return recommendations

    def getWatchlist(self):
        requestURL = f'{API.APIBASEURL}?operationName=GetWatchFavorites&variables={{}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["GetWatchFavorites"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        formats = []
        jsonData = json.loads(res)
        for element in APIParser.getElement(jsonData, ["data", "getFavoritesV3", "elements"]):
            formats.append(APIParser.parseFormatOverview(element))
        return formats

    def getFavs(self):
        requestURL = f'{API.APIBASEURL}?operationName=GetFavoriteIds&variables={{}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["GetFavoriteIds"]}"}}}}'
        res = self.getRequest(requestURL, self.token)
        jsonData = json.loads(res)
        favorites = []
        elements = APIParser.getElement(jsonData, ["data", "favorites", "elements"], True)
        if elements is not None:
            for element in elements:
                favorites.append(element['id'])
        return favorites

    def getRecommendation(self, recId, cacheFile=None):
        jsonData = {}
        formats = []
        if cacheFile is None:
            requestURL = f'{API.APIBASEURL}?operationName=ExploreWidgetWatch&variables={{"area":"home","offset":0,"take":15}}&extensions={{"persistedQuery":{{"version":1,"sha256Hash":"{API.ENDPOINTS["ExploreWidgetWatch"]}"}}}}'
            res = self.getRequest(requestURL, self.token)
            jsonData = json.loads(res)
        else:
            with open(cacheFile, encoding='utf-8') as json_file:
                jsonData = json.load(json_file)
        for result in APIParser.getElement(jsonData, ["data", "editorialView", "widgets", "items"]):
            id = APIParser.getElement(result, ["id"], True)
            if id == recId:
                elements = APIParser.getElement(result, ["elements"])
                for element in elements:
                    if element is None:
                        continue
                    formats.append(APIParser.parseFormatOverview(element))
        return formats